import numpy as np
import ray
import time
from scipy.optimize import minimize

# Define a more complex and higher-dimensional objective function for optimization
def complex_objective_function(x):
    return np.sum(np.sin(x) ** 2 + (x - 3) ** 2)  # More complex non-linear function

# Remote function to perform optimization using SciPy on each worker (Ray)
@ray.remote
def optimize_remote(initial_guess):
    start_time = time.time()  # Record start time for this worker

    # Perform optimization using SciPy's minimize function
    result = minimize(complex_objective_function, initial_guess, method='BFGS', options={'maxiter': 500})

    end_time = time.time()  # Record end time for this worker
    time_taken = end_time - start_time

    return result.x, result.fun, time_taken  # Return optimized parameters, function value, and time taken

if __name__ == "__main__":
    # Number of workers
    num_workers = 4

    # Increase dimensionality of the problem to make it more computationally expensive
    dimension = 100

    # Define different initial guesses for each worker
    initial_guesses = [np.random.randn(dimension) for _ in range(num_workers)]
    
    # ==== REMOTE EXECUTION WITH RAY ====
    ray.init()  # Initialize Ray
    print("Running optimization remotely (with Ray)...")
    overall_start_time_remote = time.time()  # Start timer for Ray execution

    # Start distributed optimization on all workers using Ray
    futures = [optimize_remote.remote(initial_guess) for initial_guess in initial_guesses]

    # Retrieve the results
    remote_results = ray.get(futures)

    overall_end_time_remote = time.time()  # End timer for Ray execution
    overall_time_remote = overall_end_time_remote - overall_start_time_remote

    # Print the results from each remote worker
    for i, (optimized_params, final_value, time_taken) in enumerate(remote_results):
        print(f"Ray Worker {i + 1} optimized parameters: {optimized_params[:5]} ... (truncated)")
        print(f"Ray Worker {i + 1} final objective value: {final_value}")
        print(f"Ray Worker {i + 1} time taken: {time_taken:.4f} seconds\n")

    print(f"Overall distributed optimization time with Ray: {overall_time_remote:.4f} seconds\n")

    # Shutdown Ray
    ray.shutdown()